@@ -3,9 +3,12 @@ from .node_classes.node import Node
 from .sec_edgar.sec_api import SecApi
 from .stocktwit import StockTwit
 
+import copy
+from datetime import datetime
 import functools
 from importlib.resources import files
 import json
+import numpy as np
 import pandas as pd
 import polars as pl
 import re
@@ -102,7 +105,7 @@ def build_rule(equation_str: str, prefer_balance: list[str] = None) -> dict:
         "prefer_balance": [v.replace(" ", "_").lower() for v in prefer_balance]
     }
 
-def triangulate_smart(df: pd.DataFrame, rules) -> (pl.DataFrame, list):
+def triangulate_smart(df: pd.DataFrame, rules) -> (pd.DataFrame, list):
     df = df.set_index("line_item")
     notes = []
     updated = set()  # Keep track of recalculated items
@@ -117,24 +120,27 @@ def triangulate_smart(df: pd.DataFrame, rules) -> (pl.DataFrame, list):
             values = [df.at[var, "value"] for var in vars_]
             var_keys = [v.replace(" ", "_").lower() for v in vars_]
 
-            # Force preferred balancing items to lowest prob
-            for i, key in enumerate(var_keys):
-                if key in prefer and vars_[i] not in updated:
-                    probs[i] = 0.0  # Ensure this becomes the target
+            # find candidates that are low confidence and not updated yet
+            low_conf_indices = [i for i, p in enumerate(probs) if p < 3.0 and vars_[i] not in updated]
 
-            low_conf_indices = [i for i, p in enumerate(probs) if p < 1.0 and vars_[i] not in updated]
+            if not low_conf_indices:
+                continue
 
-            if len(low_conf_indices) == 1:
-                i = low_conf_indices[0]
-            elif len(low_conf_indices) > 1:
-                preferred_found = [i for i in low_conf_indices if var_keys[i] in prefer]
+            # find the minimum probability among candidates
+            min_prob = min(probs[i] for i in low_conf_indices)
+            candidates = [i for i in low_conf_indices if probs[i] == min_prob]
+
+            if len(candidates) == 1:
+                i = candidates[0]
+            else:
+                # tie -> use prefer_balance if present
+                preferred_found = [j for j in candidates if vars_[j] in prefer or var_keys[j] in prefer]
                 if preferred_found:
                     i = preferred_found[0]
                 else:
-                    i = sorted(low_conf_indices, key=lambda x: probs[x])[0]
-            else:
-                continue  # All high-confidence or already updated
+                    i = candidates[0]  # fallback: pick the first
 
+            # only calculate if all other vars are known
             if all(pd.notnull(values[j]) for j in range(len(vars_)) if j != i):
                 kwargs = {var_keys[j]: None if j == i else values[j] for j in range(len(vars_))}
                 result = calc_fn(**kwargs)
@@ -225,28 +231,48 @@ class Finqual:
 
         quarter_list = self._previous_quarters(year, quarter)
 
+        if quarter_list[0][1] != 3:
+            fy_diff = self.sec_edgar.align_fy_year(False)
+        else:
+            fy_diff = 0
+
+        # ---
+
+        if "".join(label_type) in ['income_statement']:
+            fy_result = self.income_stmt(year - fy_diff, None).squeeze()
+
+        elif "".join(label_type) in ['cash_flow']:
+            fy_result = self.cash_flow(year - fy_diff, None).squeeze()
+
         quarterly_results = []
-        fy_result = self._process_financials(year, None, label_type, period_type, target_yf_list, tolerance)
 
         for period in quarter_list:
-
             curr_year = period[0]
             curr_quarter = period[1]
 
-            quarterly_results.append(self._process_financials(curr_year, curr_quarter, label_type, period_type, target_yf_list, tolerance))
+            if "".join(label_type) in ['income_statement']:
+                quarterly_results.append(self.income_stmt(curr_year, curr_quarter))
+
+            elif "".join(label_type) in ['cash_flow']:
+                quarterly_results.append(self.cash_flow(curr_year, curr_quarter))
+
+        # ---
 
         df_quarterly = pd.concat(quarterly_results, axis=1)
-        df_quarterly = df_quarterly.T.groupby(level=0).sum().T
+        df_quarterly_sum = df_quarterly.sum(axis=1)
 
-        df_annual_quarter = fy_result.copy()
-        df_annual_quarter['value'] = fy_result['value'] - df_quarterly['value']
+        df_annual_quarter = fy_result - df_quarterly_sum
+        df_annual_quarter.name = 'value'
+        df_annual_quarter = df_annual_quarter.to_frame()
 
         if "".join(label_type) in ['cash_flow']:
 
-            df_annual_quarter.loc[4, "value"] = quarterly_results[0].loc[5, "value"]  # Set Beginning Cash to last quarter End Cash
-            df_annual_quarter.loc[5, "value"] = fy_result.loc[5, "value"].round(0)  # Set End Cash to FY End Cash
+            df_annual_quarter.iloc[4, 0] = df_quarterly.iloc[5,0]  # Set Beginning Cash to last quarter End Cash
+            df_annual_quarter.iloc[5, 0] = fy_result[5].round(0)  # Set End Cash to FY End Cash
 
+        df_annual_quarter['line_item'] = df_annual_quarter.index
         df_annual_quarter['value'] = df_annual_quarter['value'].apply(lambda x: int(x))
+        df_annual_quarter['total_prob'] = 1.0
 
         return df_annual_quarter
 
@@ -258,7 +284,7 @@ class Finqual:
             tolerance = 0.0
 
         sec_data = self.sec_edgar.financial_data_period(year, quarter)
-        sec_data_dict = dict(zip(sec_data['key'], sec_data['value']))
+        sec_data_dict = dict(zip(sec_data['key'], sec_data['val']))
 
         if len(sec_data_dict) > 0:
             if quarter == self.sec_edgar.get_annual_quarter():
@@ -268,11 +294,15 @@ class Finqual:
 
             all_dfs = []
             for nodes in self.trees.values():
-                tree = NodeTree(nodes)
+                nodes_copy = copy.deepcopy(nodes)  # make a thread-safe copy
+                tree = NodeTree(nodes_copy)
+
                 tree.load_sec_data(sec_data_dict)
                 tree.get_all_values()
                 df_tree = tree.to_df()
+
                 if df_tree is not None:
+                    df_tree = df_tree.with_columns(pl.col("balance").cast(pl.Utf8))
                     all_dfs.append(df_tree)
 
             df_total = pl.concat(all_dfs, how="vertical") if all_dfs else pl.DataFrame()
@@ -287,7 +317,7 @@ class Finqual:
 
             df_total = df_total.select(['yf', 'prob', 'value'])
             df_total = (df_total.group_by(['yf', 'value']).agg(pl.col('prob').sum().alias('total_prob')).sort('total_prob', descending=True))
-            df_total = (df_total.sort(['yf', 'total_prob'], descending=[False, True]).unique(subset='yf', keep='first'))
+            df_total = (df_total.sort(['yf', 'total_prob', 'value'], descending=[False, True, False]).unique(subset='yf', keep='first'))
             df_total = df_total.filter(pl.col('total_prob') >= tolerance)
 
             # ---
@@ -306,9 +336,10 @@ class Finqual:
             return df_total.to_pandas()
 
         else:
-            print(f"*** Finqual: There is no data available for ticker {self.ticker} - it may be too newly listed or in the future.")
+            #print(f"*** Finqual: There is no data available for ticker {self.ticker} for year {year} and/or quarter {quarter} - it may be too newly listed or in the future. \n")
             df_target = pl.DataFrame({"line_item": target_yf_list})
             df_target = df_target.with_columns(pl.lit(0).alias("value"))
+            df_target = df_target.with_columns(pl.lit(0).alias("total_prob"))
 
             return df_target.to_pandas()
 
@@ -324,7 +355,7 @@ class Finqual:
         ]
 
         rules = [
-            build_rule("Gross Profit = Total Revenue - Cost Of Revenue"),
+            build_rule("Gross Profit = Total Revenue - Cost Of Revenue", prefer_balance=["Gross Profit"]),
             build_rule("Operating Income = Gross Profit - Selling General And Administration - Research And Development - Other Operating Income Expense", prefer_balance=["Other Operating Income Expense"]),
             build_rule("Pretax Income = Operating Income - Interest Expense + Other Non Operating Income Expense", prefer_balance=["Other Non Operating Income Expense"]),
             build_rule("Net Income = Pretax Income - Tax Provision"),
@@ -338,6 +369,23 @@ class Finqual:
             target_yf_list=tuple(income_list),
         )
 
+        # ---
+
+        df_income["total_prob"] = df_income["total_prob"].astype(float)
+
+        increments = {
+            "Total Revenue": 0.5,
+            "Gross Profit": 0.5,
+            "Operating Income": 0.5,
+            "Pretax Income": 0.5,
+            "Net Income": 0.5
+        }
+
+        for item, inc in increments.items():
+            df_income.loc[(df_income["line_item"] == item) & (df_income["total_prob"] != 0), "total_prob"] += inc
+
+        # ---
+
         df_income, log = triangulate_smart(df_income, rules)
 
         label = str(year) if quarter is None else str(year) + "Q" + str(quarter)
@@ -364,7 +412,7 @@ class Finqual:
         df_bs = self._process_financials(
             year, quarter,
             label_type=tuple(["balance_sheet"]),
-            period_type=tuple(["instant", "duration"]),
+            period_type=tuple(["instant"]),
             target_yf_list=tuple(balance_list),
         )
 
@@ -430,7 +478,16 @@ class Finqual:
         return df_cf
 
     def _financials_period(self, method_name: str, start_year: int, end_year: int, append_type: str, quarter: bool = False):
+        """
+        General method to retrieve financial statements or ratios over a given period
 
+        :param method_name: Method name within Finqual to call
+        :param start_year: Start of period
+        :param end_year: End of period
+        :param append_type: Tells the method whether it is a 'statement' or 'ratios' to look for
+        :param quarter: Returns quarterly info
+        :return:
+        """
         func = getattr(self, method_name)
 
         years_period = [i for i in range(end_year, start_year - 1, -1)]
@@ -450,6 +507,14 @@ class Finqual:
                 df = future.result()
                 results[label] = df
 
+        # # - This is sequential and confirmed to work
+        # for y in years_period:
+        #     if not quarter:
+        #         results[f"{y}"] = func(y)
+        #     else:
+        #         for q in [4, 3, 2, 1]:
+        #             results[f"{y}Q{q}"] = func(y, q)
+
         if append_type == 'statement':
 
             if quarter:
@@ -458,8 +523,8 @@ class Finqual:
                 ordered_labels = [f"{y}" for y in years_period]
 
             # Ordering and formatting the result
-            ordered_dfs = [results[label] for label in ordered_labels if label in results]
-            df_total = pd.concat(ordered_dfs, axis=1)
+            ordered_results = [results[label] for label in ordered_labels]
+            df_total = pd.concat(ordered_results, axis=1)
 
             df_total = df_total.loc[:, (df_total != 0).any(axis=0)]
             df_total = df_total[[label for label in ordered_labels if label in df_total.columns]]
@@ -482,19 +547,91 @@ class Finqual:
         return self._financials_period('income_stmt', start_year, end_year, 'statement', quarter)
 
     def balance_sheet_period(self, start_year: int, end_year: int, quarter: bool = False):
-        return self._financials_period('balance_sheet', start_year, end_year, 'statement', quarter)
+
+        df_bs = self._financials_period('balance_sheet', start_year, end_year, 'statement', quarter)
+        cols_to_drop = [col for col in df_bs.columns if (df_bs.drop(index="Shares Outstanding")[col] == 0).all()]
+        df_bs = df_bs.drop(columns=cols_to_drop)
+
+        return df_bs
 
     def cash_flow_period(self, start_year: int, end_year: int, quarter: bool = False):
         return self._financials_period('cash_flow', start_year, end_year, 'statement', quarter)
 
-    def _get_ratios(self, year: int, ratio_definitions: dict, statement_fetchers: dict, quarter: int | None = None, pct_flag: bool = False):
+    def income_stmt_ttm(self):
+        current_year = int(datetime.now().year)
+        df_inc = self.income_stmt_period(current_year - 2, current_year + 1, True)
 
+        if len(df_inc.columns) < 4:
+            print("Not enough data to calculate TTM")
+            df_ttm = pd.DataFrame(df_inc[df_inc.columns[:4]].sum(axis=1), columns = ["TTM"])
+            df_ttm["TTM"] = "Not Found"
+            return df_ttm
+
+        df_ttm = pd.DataFrame(df_inc[df_inc.columns[:4]].sum(axis=1), columns = ["TTM"])
+        return df_ttm
+
+    def balance_sheet_ttm(self):
+        current_year = int(datetime.now().year)
+        df_bs = self.balance_sheet_period(current_year - 1, current_year + 1, True)
+
+        if len(df_bs.columns) < 1:
+            print("Not enough data to calculate latest balance sheet.")
+            df_ttm = pd.DataFrame(df_bs[df_bs.columns[:1]].sum(axis=1), columns = ["TTM"])
+            df_ttm["TTM"] = "Not Found"
+            return df_ttm
+
+        df_ttm = pd.DataFrame(df_bs[df_bs.columns[:1]].sum(axis=1), columns = ["TTM"])
+
+        return df_ttm
+
+    def cash_flow_ttm(self):
+        current_year = int(datetime.now().year)
+        df_cf = self.cash_flow_period(current_year - 2, current_year + 1, True)
+
+        if len(df_cf.columns) <= 4:
+            print("Not enough data to calculate latest TTM for cashflow.")
+            df_ttm = pd.DataFrame(df_cf[df_cf.columns[:4]].sum(axis=1), columns = ["TTM"])
+            df_ttm["TTM"] = "Not Found"
+            return df_ttm
+
+        df_ttm = df_cf[df_cf.columns[:4]].copy()
+
+        # ---
+
+        df_ttm["TTM"] = df_ttm.iloc[:4,:].sum(axis=1)
+
+        df_ttm.loc["Beginning Cash Position", "TTM"] = df_cf.iloc[4, 3]
+        df_ttm.loc["End Cash Position", "TTM"] = df_cf.iloc[5, 0]
+        df_ttm.loc["Changes In Cash", "TTM"] = df_cf.iloc[5, 0] - df_cf.iloc[4, 3]
+
+        df_ttm = pd.DataFrame(df_ttm["TTM"], columns=["TTM"])
+
+
+        return df_ttm
+
+    # ----
+
+    def _get_ratios(self, year: int|None, ratio_definitions: dict, statement_fetchers: dict, quarter: int | None = None, pct_flag: bool = False):
+        """
+        Method that takes in a dictionary with definitions of ratios and a dictionary of statements to retrieve (i.e. income, balance or cashflow)
+
+        :param year: The year of info to retrieve from
+        :param ratio_definitions: Dictionary of ratio definitions/formula
+        :param statement_fetchers: Dictionary of statements to retrieve (i.e. income, balance, cashflow)
+        :param quarter: The specific quarter to look at
+        :param pct_flag: True/False for whether the ratio is expressed in terms of percentages
+        :return:
+        """
         if pct_flag:
             ratio_multiplier = 100
         else:
             ratio_multiplier = 1
 
-        label = str(year) + "Q" + str(quarter) if quarter is not None else str(year)
+        if year is None and quarter is None:
+            label = "TTM"
+
+        else:
+            label = str(year) + "Q" + str(quarter) if quarter is not None else str(year)
 
         try:
             statement_data = {}
@@ -513,6 +650,7 @@ class Finqual:
                 except (ZeroDivisionError, TypeError):
                     result[ratio] = "Not Found"
             return result
+
         except KeyError:
             print(f"No data for {self.ticker} found.")
             result = {'Ticker': self.ticker, 'Period': label}
@@ -559,9 +697,35 @@ class Finqual:
 
         return df_ratio
 
-    def valuation_ratios(self, year: int, quarter: int | None = None):
+    def valuation_ratios(self, year: int|None = None, quarter: int | None = None):
+        """
+        Retrieve valuation for a given time period - only works for latest time.
 
-        share_price = StockTwit(self.ticker).retrieve_data()[self.ticker]
+        :param year: Year of interest
+        :param quarter: Quarter of interest
+        :return:
+        """
+        if year is None and quarter is None:
+
+            share_price = StockTwit(self.ticker).retrieve_data()[self.ticker]  # This would have to be the share price at the given year and quarter time
+
+            statement_fetchers = {
+                'income': lambda y, q=None: self.income_stmt_ttm(),
+                'balance': lambda y, q=None: self.balance_sheet_ttm(),
+                'cashflow': lambda y, q=None: self.cash_flow_ttm(),
+            }
+
+        else:
+
+            print("*** Finqual: Note that functionality for historical valuation ratios not implemented.")
+
+            share_price = np.nan  # Placeholder code - need to add the share price at the requested year and quarter
+
+            statement_fetchers = {
+            'income': lambda y, q=None: self.income_stmt(y, q) if q else self.income_stmt(y),
+            'balance': lambda y, q=None: self.balance_sheet(y, q) if q else self.balance_sheet(y),
+            'cashflow': lambda y, q=None: self.cash_flow(y, q) if q else self.cash_flow(y),
+            }
 
         ratio_definitions = {
             'EPS': lambda data: data['income'].get('Net Income') / data['balance'].get('Shares Outstanding'),
@@ -570,12 +734,6 @@ class Finqual:
             'EV/EBITDA': lambda data: (data['balance'].get('Shares Outstanding') * share_price + data['cashflow'].get('End Cash Position')) / (data['income'].get('Operating Income') + data['cashflow'].get('Depreciation And Amortization')),
         }
 
-        statement_fetchers = {
-            'income': lambda y, q=None: self.income_stmt(y, q) if q else self.income_stmt(y),
-            'balance': lambda y, q=None: self.balance_sheet(y, q) if q else self.balance_sheet(y),
-            'cashflow': lambda y, q=None: self.cash_flow(y, q) if q else self.cash_flow(y),
-        }
-
         ratios = self._get_ratios(year, ratio_definitions, statement_fetchers, quarter, False)
         df_ratio = pd.DataFrame([ratios])
         df_ratio = df_ratio.fillna("Not Found")
